#ifndef APACHE_MULTITHREAD_H
#define APACHE_MULTITHREAD_H

#define MULTI_OK (0)
#define MULTI_TIMEOUT (1)
#define MULTI_ERR (2)

typedef void mutex;

/*
 * Ambarish: Need to do the right stuff on multi-threaded unix
 * I believe this is terribly ugly
 */
#ifdef MULTITHREAD
typedef void semaphore;
typedef void thread;
typedef void event;
#define APACHE_TLS __declspec( thread )

thread *create_thread(void (thread_fn) (void *thread_arg), void *thread_arg);
int kill_thread(thread *thread_id);
int await_thread(thread *thread_id, int sec_to_wait);
void exit_thread(int status);
void free_thread(thread *thread_id);

API_EXPORT(mutex *) ap_create_mutex(char *name);
API_EXPORT(mutex *) ap_open_mutex(char *name);
API_EXPORT(int) ap_acquire_mutex(mutex *mutex_id);
API_EXPORT(int) ap_release_mutex(mutex *mutex_id);
API_EXPORT(void) ap_destroy_mutex(mutex *mutex_id);

semaphore *create_semaphore(int initial);
int acquire_semaphore(semaphore *semaphore_id);
int release_semaphore(semaphore *semaphore_id);
void destroy_semaphore(semaphore *semaphore_id);

event *create_event(int manual, int initial, char *name);
event *open_event(char *name);
int acquire_event(event *event_id);
int set_event(event *event_id);
int reset_event(event *event_id);
void destroy_event(event *event_id);

#elif defined(NSPR)

#include <prlock.h>

extern inline mutex *ap_create_mutex(char *name);
extern inline mutex *ap_create_mutex(char *name)
{
    PRLock *l;

    l = PR_NewLock();
    if (l) {
	return l;
    }
    fprintf(stderr, "error from PR_NewLock\n");
    abort();
}

extern inline int ap_acquire_mutex(mutex *vmx);
extern inline int ap_acquire_mutex(mutex *vmx)
{
    PRLock *l = vmx;

    PR_Lock(l);
    return MULTI_OK;
}

extern inline int ap_release_mutex(mutex *vmx);
extern inline int ap_release_mutex(mutex *vmx)
{
    PRLock *l = vmx;
    PRStatus s;

    s = PR_Unlock(l);
    return (s == PR_SUCCESS ? MULTI_OK : MULTI_ERR);
}

extern inline int ap_destroy_mutex(mutex *vmx);
extern inline int ap_destroy_mutex(mutex *vmx)
{
    PRLock *l = vmx;

    PR_DestroyLock(l);
}

#elif defined(PTHREAD)

typedef struct {
    pthread_mutex_t px;
    pid_t pid;
} lthread_mutex_t;

extern inline mutex *ap_create_mutex(char *name);
extern inline mutex *ap_create_mutex(char *name)
{
    static const pthread_mutex_t def = PTHREAD_MUTEX_INITIALIZER;
    lthread_mutex_t *mx;
    int rc;

    mx = malloc(sizeof(*mx));
    if (mx == NULL) {
	fprintf(stderr, "out of memory!\n");
	abort();
    }

    mx->px = def;
    mx->pid = 0;
    if ((rc = pthread_mutex_init(&mx->px, NULL))) {
	fprintf(stderr, "pthread_mutex_init: %s\n", strerror(rc));
	abort();
    }

    return mx;
}

extern inline int ap_acquire_mutex(mutex *vmx);
extern inline int ap_acquire_mutex(mutex *vmx)
{
    lthread_mutex_t *mx = vmx;
    pid_t me = getpid();

    if (mx->pid == me) {
	return MULTI_OK;
    }
    if (pthread_mutex_lock(&mx->px)) {
	return MULTI_ERR;
    }
    mx->pid = me;
    return MULTI_OK;
}

extern inline int ap_release_mutex(mutex *vmx);
extern inline int ap_release_mutex(mutex *vmx)
{
    lthread_mutex_t *mx = vmx;

    mx->pid = 0;
    if (pthread_mutex_unlock(&mx->px)) {
	return MULTI_ERR;
    }
    return MULTI_OK;
}

extern inline int ap_destroy_mutex(mutex *vmx);
extern inline int ap_destroy_mutex(mutex *vmx)
{
    lthread_mutex_t *mx = vmx;
    pthread_mutex_destroy(&mx->px);
    free(mx);
}

#else /* ndef MULTITHREAD */

#define APACHE_TLS
/* Only define the ones actually used, for now */
extern void *ap_dummy_mutex;

#define ap_create_mutex(name)	((mutex *)ap_dummy_mutex)
#define ap_acquire_mutex(mutex_id)	((int)MULTI_OK)
#define ap_release_mutex(mutex_id)	((int)MULTI_OK)
#define ap_destroy_mutex(mutex_id)	(0)

#endif /* ndef MULTITHREAD */

#endif /* !APACHE_MULTITHREAD_H */
